import numpy as np
import hgdl.misc as misc
import hgdl.local_methods.bump_function as defl
import dask.distributed as distributed


def DNewton(func,grad,hess,bounds,x0,max_iter,*args):
    e = np.inf
    tol = 1e-6
    counter = 0
    x = np.array(x0)
    success = True
    while e > tol:
        counter += 1
        gradient = grad(x,*args)
        hessian  = hess(x,*args)
        e = np.linalg.norm(gradient)
        try:
            gamma = np.linalg.solve(hessian,-gradient)
        except Exception as error: 
            gamma,a,b,c = np.linalg.lstsq(hessian,-gradient)
        x += gamma
        if misc.out_of_bounds(x,bounds):
            x = np.random.uniform(low = bounds[:,0], high = bounds[:,1], size = len(bounds))
        if counter > max_iter: return x,func(x, *args),e,np.linalg.eig(hess(x, *args))[0], False
    return x,func(x, *args),e,np.linalg.eig(hess(x, *args))[0], success



def gradient_descent(ObjectiveFunction, GradientFunction,bounds,x0, radius, args):
    dim = len(bounds)
    bounds = np.array(bounds)
    epsilon = np.inf
    step_counter = 0
    beta = 0.5
    x = np.array(x0)
    success = True
    while epsilon > 1e-6:
        step_counter += 1
        aug_gradient = defl.deflated_grad(GradientFunction,x,x_defl,radius, *args)
        counter = 0
        step = 1.0
        while misc.out_of_bounds(x - (step * aug_gradient), bounds) or \
              ObjectiveFunction(x - (step * aug_gradient), *args) > \
              ObjectiveFunction(x, *args) - ((step / 2.0) * np.linalg.norm(aug_gradient) ** 2):
            step = step * beta
            counter += 1
            if counter > 10:
                return x, ObjectiveFunction(x, *args), False
        x = x - (step * aug_gradient)
        epsilon = np.linalg.norm(aug_gradient)
        if step_counter > 20:
            success = False
            break
    return x, ObjectiveFunction(x, *args), True




import numpy as np


class meta_data:
    """
    this class contains all the data necessary to run the workers
    """
    def __init__(self,obj):
        self.workers = obj.workers    ###dictionary oh ost and walker workers
        self.func = obj.func
        self.grad = obj.grad
        self.hess = obj.hess
        self.bounds = obj.bounds
        self.radius = obj.radius
        self.dim = obj.dim
        self.global_tol = obj.global_tol
        self.local_max_iter = obj.local_max_iter
        self.number_of_walkers = obj.number_of_walkers
        self.num_epochs = obj.num_epochs
        self.global_optimizer = obj.global_optimizer
        self.local_optimizer = obj.local_optimizer
        self.args = obj.args
        self.constr = obj.constr



##global optimizer for hgdl
import numpy as np
import hgdl.misc as misc


def run_global(x,y,bounds,method,number_of_offspring):
    if method == "genetic": return genetic_step(x,y,bounds,number_of_offspring)
    elif method == "gauss": return gauss_step(x,y,bounds,number_of_offspring)
    elif method == "bayes": return bayes_step(x,y,bounds,number_of_offspring)
    elif method is callable: return method(x,y,bounds,number_of_offspring)
    else: raise Exception("no global method specified")

def gauss_step(x, y, bounds,n):
    ####x is x where in bounds
    x_new = []
    y_new = []
    #bounds = np.array(bounds)
    for i in range(len(x)):
        if misc.in_bounds(x[i],bounds): 
            x_new.append(x[i])
            y_new.append(y[i])
    x = np.asarray(x_new)
    y = np.asarray(y_new)
    sorted_indices = np.argsort(y)
    y = y[sorted_indices]
    x = x[sorted_indices]
    x = np.array(x[:-int(len(x)/2)])
    y = np.array(y[:-int(len(x)/2)])
    cov = np.cov(x.T)
    mean= np.mean(x , axis = 0)
    offspring = np.random.multivariate_normal(mean, cov, size = n)
    for i in range(len(offspring)):
        if not misc.in_bounds(offspring[i],bounds): 
            offspring[i] = np.random.uniform(low = bounds[:,0],high = bounds[:,1], size = len(offspring[0]))
    return offspring
###########################################################################
def genetic_step(X, y, bounds, numChoose):
    """
    Input:
    X is the individuals - points on a surface
    y is the performance - f(X)
    Notes:
    the children can be outside of the bounds!
    """
    unfairness = 2.5
    wildness = 0.05
    N, k = X.shape
    # normalize the performances to (0,1)
    y -= np.amin(y)
    amax = np.amax(y)
    # if the distribution of performance has no width,
    #   give everyone an equal shot
    if np.isclose(amax,0.):
        p = np.ones(N)*1./N
    else:
        y /= np.amax(y)
        y *= -1.
        y -= np.amin(y)
        y += 1
        p = y/np.sum(y)
    #This chooses from the sample based on the power law,
    #allowing replacement means that the the individuals
    #can have multiple kids
    p = unfairness*np.power(p,unfairness-1)
    p /= np.sum(p)
    if np.isnan(p).any():
        raise Exception("got isnans in GeneticStep")
    moms = np.random.choice(N, size=numChoose, replace=True, p=p)
    dads = np.random.choice(N, size=numChoose, replace=True, p=p)
    # calculate a perturbation to the median
    #   of each individual's parents
    perturbation = np.random.normal(
            loc = 0.,
            scale=wildness*(bounds[:,1]-bounds[:,0]),
            size=(numChoose,k))
    # the children are the median of their parents plus a perturbation
    norm = p[moms]+p[dads]
    weights = (p[moms]/norm, p[dads]/norm)
    weighted_linear_sum = weights[0].reshape(-1,1)*X[moms] + weights[0].reshape(-1,1)*X[dads]
    children = weighted_linear_sum + perturbation
    oob = np.logical_not([misc.in_bounds(x,bounds) for x in children])
    children[oob] = misc.random_sample(np.sum(oob), k, bounds)
    return children


def bayes_step(x,y,bounds, number_of_offspring):
    raise Exception("bayesian global step no yet implemented")



###local optimizer for hgdl
import numpy as np
import hgdl.misc as misc
import asyncio
import time
import matplotlib.pyplot as plt
def b(x,x0,r):
    """
    evaluates the bump function
    x ... a point (1d numpy array)
    x0 ... 1d numpy array of location of bump function
    returns the bump function b(x,x0) with radius r
    """
    d = np.sqrt((x - x0).T @ (x - x0))
    a = 1.0-(d**2/r**2)
    if a <= 0: return 0.0
    else: return np.exp(-1.0/a) * np.exp(1.0)
###########################################################################
def b_grad(x,x0,r):
    """return the bump function gradient b(x,x0) with radius r"""
    d = np.sqrt((x - x0).T @ (x - x0))
    d2= (x - x0)
    gr = np.zeros((d2.shape))
    a = 1.0-(d**2/r**2)
    if a <= 0: return np.zeros((len(x)))
    else: return (b(x,x0,r) * ((-2.0*d2)/(a**2))) /r**2
###########################################################################
def deflation_function(x,x0,r):
    """
    input:
        x is one point(1d numpy array)
        x0 is a a 2d array of locations of the bump function
    the return is the deflation operator, e.g. 1.0/(1.0 - bump(x,x0))
    """
    if len(x0) == 0: s = 0.0; return (1.0/(1.0-s))
    s = 0.0
    for i in range(len(x0)):
        s += b(x,x0[i],r)
    return (1.0/(1.0-s))
###########################################################################
def deflation_function_gradient(x,x0,r):
    """
    input:
        x is one point(1d numpy array)
        x0 is a a 2d array of locations of the bump function
    the return is the gradient of the deflation operator, e.g. (1.0/(1.0 - bump(x,x0)))'
    """
    if len(x0) == 0: return np.zeros((len(x)))
    s1 = 0.0
    s2 = np.zeros((len(x)))
    for i in range(len(x0)):
        s1 += b(x,x0[i],r)
        s2 += b_grad(x,x0[i],r)
    return (1.0/((1.0-s1)**2)) * s2


def deflated_hess(x,*args, grad_func = None, hess_func = None, x_defl = [], 
                  radius = 0.5, extended_return = False):
    d = deflation_function(x,x_defl,radius)
    dg = deflation_function_gradient(x,x_defl,radius)
    if extended_return == True: return (hess_func(x, *args)/d) + (np.outer(grad_func(x, *args),dg)/d**2),d,dg
    return (hess_func(x, *args)/d) + (np.outer(grad_func(x, *args),dg)/d**2)



def deflated_solve( x, *args, grad_func = None, hess_func = None, x_defl=[],
                    radius = 0.5, extended_return = False):
    d = deflation_function(x,x_defl,radius)
    dg = deflation_function_gradient(x,x_defl,radius)
    if extended_return == True: return hess_func(x, *args) + (np.outer(grad_func(x, *args),dg)/d),d,dg
    return hess_func(x, *args)+ (np.outer(grad_func(x, *args),dg)/d)

def deflated_grad(x, *args, grad_func = None, x_defl = [], radius = 0.5):
    d = deflation_function(x,x_defl,radius)
    return d*grad_func(x, *args)

